
#include "TAnalysis14Be.hh"
#include "TAnalysisTools.hh"

#include "TFile.h"
#include "TDirectory.h"
#include "TTree.h"
#include "TH1D.h"
#include "TH2D.h"
#include "TRandom.h"
#include "TVector3.h"

#include <iostream>
#include <sstream>
#include <string>
#include <cmath>
#include <vector>

#include "TSimulationDataBeam.hh"
#include "TSimulationParameterGlobal.hh"
#include "TSimulationDataScintillator.hh"
#include "TSimulationParameterNeutronDetector.hh"

ClassImp(TAnalysis14Be)

TAnalysis14Be::TAnalysis14Be()
: TAnalysisNebula()
{;}

TAnalysis14Be::~TAnalysis14Be()
{;}

int TAnalysis14Be::ReadParameters(TFile* file)
{
  TAnalysisNebula::ReadParameters(file);

  fNeutHardwareThreshold = 2; // MeVee
  fVetoHardwareThreshold = 0.1; // MeVee
  fNeutSoftwareThreshold = 6; // MeVee
  fVetoSoftwareThreshold = 1; // MeVee

  fTimeResolution = 0.17*sqrt(2); // ns
  //  fTimeResolution = 1; // ns

  fNeutAttenuationLength = 668; // cm
  fVetoAttenuationLength = 258; // cm
  fPropagationVelocity = 15.8; // cm/ns

  return 1;
}

int TAnalysis14Be::CreateHist()
{
  TAnalysisNebula::CreateHist();

  fHErel = new TH1D("herel", "erel", 800, 0, 40);
  fHErelBetaDBeta = new TH2D("herelbetadbeta", "erel vs betadbeta", 800, 0, 40, 200, -5, 5);

  return 0;
}

int TAnalysis14Be::ProcessOneEvent()
{
  if(gSimulationDataScintillatorArray->empty()){
#ifdef DEBUG
    std::cout << " no events" << std::endl;
#endif
    return 0;
  }

  // --------------------- create experiment data ---------------------
  fID.clear();
  fMT.clear();
  fQL.clear();
  fQR.clear();
  fTL.clear();
  fTR.clear();
  SimToExp();

  // ---------------------- fill data container --------------------------
  std::vector<TAnalysisNebulaData> data_v;
  for(int j=0; j<(int)fID.size(); ++j){
    TAnalysisNebulaData data;
    std::vector<int> id = fID[j];
    data.id = fID[j];
    data.sid = -1;
    if(fMT[j] == kNEUT){
      data.sid = 1+id[0]+id[1]*fSimParaNeut->fNeutNumber+id[2]*fSimParaNeut->fNeutNumber*fSimParaNeut->fNeutLayerNumber;
    }else if(fMT[j] == kVETO){
      data.sid = 1+id[0]+id[2]*fSimParaNeut->fVetoNumber+fTotalNeutNumber;
    }
    data.mt = fMT[j];
    data.qu = fQL[j];
    data.qd = fQR[j];
    data.tu = fTL[j];
    data.td = fTR[j];
    if(data.mt == kNEUT){
      data.qave = sqrt(data.qu*data.qd)*exp(fSimParaNeut->fNeutSize.X()/(2*fNeutAttenuationLength));
      data.tave = (data.tu+data.td)/2-fSimParaNeut->fNeutSize.X()/fPropagationVelocity/2;
    }else{
      data.qave = sqrt(data.qu*data.qd)*exp(fSimParaNeut->fVetoSize.X()/(2*fVetoAttenuationLength));
      data.tave = (data.tu+data.td)/2-fSimParaNeut->fVetoSize.X()/fPropagationVelocity/2;
    }	
    data.dt = -(data.tu-data.td);

    if(data.mt == kNEUT){
      data.pos = fSimParaNeut->fNeutPos + fSimParaNeut->fNeutPosShift*id[0];
      data.pos += fSimParaNeut->fNeutLayerPos + fSimParaNeut->fNeutLayerPosShift*id[1];
    }else{
      data.pos.SetX(fSimParaNeut->fVetoPosShift.x()*id[0]);
      data.pos.SetY(fSimParaNeut->fVetoPosShift.y()*id[0]);
      if(id[0]%2){
	data.pos.SetZ(-1*fSimParaNeut->fVetoPosShift.z());
      }
      data.pos += fSimParaNeut->fVetoPos;
      data.pos += fSimParaNeut->fVetoLayerPos;
    }
    data.pos += fSimParaNeut->fStuckPos + fSimParaNeut->fStuckPosShift*id[2];
    data.pos.SetX(data.pos.x()+0.01+data.dt*fPropagationVelocity/2);
    data.pos.SetY(data.pos.y()+0.01);
    data.pos.SetZ(data.pos.z()-fSimParaNeut->fNeutFaceZ);
    //    data.pos.SetY(data.pos.y()+0.01+gRandom->Uniform(fSimParaNeut->fNeutSize.Y())-fSimParaNeut->fNeutSize.Y()/2);
    //    data.pos.SetZ(data.pos.z()-fSimParaNeut->fNeutFaceZ+gRandom->Uniform(fSimParaNeut->fNeutSize.Z())-fSimParaNeut->fNeutSize.Z()/2);

    if(fSimParaNeut->fNeutRotationMode == 1){
      double y = data.pos.x();
      double x = data.pos.y();
      data.pos.SetX(x);
      data.pos.SetY(y);
    }else if(fSimParaNeut->fNeutRotationMode == 2){
      if(id[1]%2){
	double y = data.pos.x();
	double x = data.pos.y();
	data.pos.SetX(x);
	data.pos.SetY(y);
      }
    }
     
    // add overall offset
    data.pos += fSimParaNeut->fPosition;

    // to adjust to experiment
    data.pos.SetX(-data.pos.x());

    //    std::cout << data.pos.x() << " " << data.pos.y() << " " << data.pos.z() << " " << fSimParaNeut->fPosition.z() << std::endl;

    Double_t time = data.tave;
    TVector3 beta, p;
    for(Int_t i=0; i<3; ++i){
      beta[i] = data.pos[i]/time/vlight;
    }
    Double_t betaabs = sqrt(beta[0]*beta[0]+beta[1]*beta[1]+beta[2]*beta[2]);
    Double_t gamma = 1/sqrt(1-betaabs*betaabs);
    for(Int_t i=0; i<3; ++i){
      p[i] = mneutron*gamma*beta[i];
    }
    data.time = time;
    data.betaabs = betaabs;
    data.pabs = sqrt(p[0]*p[0]+p[1]*p[1]+p[2]*p[2]);
    data.energy = sqrt(mneutron*mneutron + data.pabs*data.pabs) - mneutron;
    data.theta = atan(sqrt(p[0]*p[0]+p[1]*p[2])/p[2]);
    data.phi = atan(p[1]/p[0]);

    data.beta = beta;
    data.gamma = gamma;
    data.p = p;

    //    std::cout << j << " " << data.sid << " " << data.time << " " << data.pos[0] << " " << data.pos[1] << " " << data.pos[2] << " " << time << " " << vlight << " " << beta[0] << " " << beta[1] << " " << beta[2] << " " << data.betaabs << " " << data.pabs << " " << data.energy << " " << data.theta << std::endl;

    data_v.push_back(data);
  }

  if(data_v.empty()) return 1;

  //  std::cout << gSimulationDataBeamArray->size() << std::endl;
  TSimulationDataBeam dataBeam = (*gSimulationDataBeamArray)[0];
  TVector3 beammomentum = dataBeam.fMomentum;
  double beamp = dataBeam.fMomentum.Mag();
  double beamm = dataBeam.fMass;
  double beammovp = beamm/beamp;
  double beambeta = 1/sqrt(beammovp*beammovp+1);
  double beamenergy = sqrt(beamm*beamm+beamp*beamp);
  double beamk = beamenergy - beamm;

//  // --- Erel ---
//  TVector3 ptot;
//  Double_t etot=0;
//  Double_t mtot=0;
//  for(int i=0; i<3; ++i){
//    TSimulationDataBeam dataBeam = (*gSimulationDataBeamArray)[i];
//    ptot += dataBeam.fMomentum;
//    etot += sqrt(pow(dataBeam.fMomentum.Mag(),2)+pow(dataBeam.fMass,2));
//    mtot += dataBeam.fMass;
//  }
//  //  Double_t erel = sqrt(etot*etot - ptot*ptot)-mtot;
//  //  std::cout << erel << std::endl;

  // -------------------------- analyze data -------------------------------------
  TAnalysisNebulaTools::threshold(data_v, fNeutSoftwareThreshold, fVetoSoftwareThreshold);
  if(data_v.empty()) return 1;

  if(fNumberOfNeutrons == 1){
    if(TAnalysisNebulaTools::veto(data_v, fVetoMode)) return 1;
    TAnalysisNebulaTools::killveto(data_v);
    if(data_v.empty()) return 1;

    fHMulti->Fill(data_v.size());
    for(int j=0; j<(int)data_v.size(); ++j){
      fHIDQAveAll->Fill(data_v[j].sid, data_v[j].qave);
    }
    TAnalysisNebulaTools::hitmintime(data_v);
  }else if(fNumberOfNeutrons == 2){ // search cross talk
    // select 2hit
    //    if(TAnalysisNebulaTools::veto(data_v, fVetoMode)) return 1;
    if(TAnalysis14Be::vetolocal(data_v)) return 1;
    TAnalysisNebulaTools::killveto(data_v);
    fHMulti->Fill(data_v.size());
    for(int j=0; j<(int)data_v.size(); ++j){
      fHIDQAveAll->Fill(data_v[j].sid, data_v[j].qave);
    }
    TAnalysisNebulaTools::hitmintime2(data_v);
    if(data_v.size() != 2) return 1;
  }   

  // cross talk fill
  if(fNumberOfNeutrons == 2){
    TAnalysisNebulaData n1 = data_v[0];
    TAnalysisNebulaData n2 = data_v[1];
    if(data_v[0].pos[2] > data_v[1].pos[2]){
      n1 = data_v[1];
      n2 = data_v[0];
    }else if(data_v[0].pos[2] == data_v[1].pos[2]){
      if(data_v[0].time > data_v[1].time){
	n1 = data_v[1];
	n2 = data_v[0];
      }
    }
    double t1 = n1.time;
    double t2 = n2.time;
    double q1 = n1.qave;
    double q2 = n2.qave;
    TVector3 pos1 = n1.pos;
    TVector3 pos2 = n2.pos;

    double x1 = pos1[0];
    double y1 = pos1[1];
    double z1 = pos1[2];
    double x2 = pos2[0];
    double y2 = pos2[1];
    double z2 = pos2[2];
    double dx = x2-x1;
    double dy = y2-y1;
    double dz = z2-z1;
    double dr = sqrt(dx*dx+dy*dy+dz*dz);
    double dt = t2-t1;
    double dbeta = dr/dt/vlight;
    //    double betaddbeta = (t1 < t2) ? dbeta/n1.betaabs : dbeta/n2.betaabs;
    double betaddbeta = dbeta/n1.betaabs;
    double ivbetaddbeta = 1/betaddbeta;

    TVector3 np1 = n1.p;
    TVector3 np2 = n2.p;
    TVector3 nbeta1 = n1.beta;
    TVector3 nbeta2 = n2.beta;
    double ne1 = n1.energy;
    double ne2 = n2.energy;

    fhq1q2->Fill(q1,q2);

    fhdtq1->Fill(dt, q1);
    fhdtq2->Fill(dt, q2);
    fhdbetaq1->Fill(dbeta, q1);
    fhdbetaq2->Fill(dbeta, q2);
    fhivdbetaq1->Fill(1/dbeta, q1);
    fhivdbetaq2->Fill(1/dbeta, q2);
    fhbetaddbetaq1->Fill(betaddbeta, q1);
    fhbetaddbetaq2->Fill(betaddbeta, q2);
    fhivbetaddbetaq1->Fill(ivbetaddbeta, q1);
    fhivbetaddbetaq2->Fill(ivbetaddbeta, q2);

    // erel
    TVector3 ptot;
    Double_t etot=0;
    Double_t mtot=0;
    {
      TSimulationDataBeam dataBeam = (*gSimulationDataBeamArray)[2];
      ptot = dataBeam.fMomentum+np1+np2;
      etot = sqrt(pow(dataBeam.fMomentum.Mag(),2)+pow(dataBeam.fMass,2))+ne1+ne2+2*mneutron;
      mtot = dataBeam.fMass+2*mneutron;
    }
    Double_t erel = sqrt(etot*etot - ptot*ptot)-mtot;
    fHErel->Fill(erel);
    fHErelBetaDBeta->Fill(erel, ivbetaddbeta);
    //    std::cout << erel << std::endl;
  }

  // 1n fill
  for(int j=0; j<(int)data_v.size(); ++j){
    Int_t id = data_v[j].sid;
    Double_t qu = data_v[j].qu;
    Double_t qd = data_v[j].qd;
    Double_t tu = data_v[j].tu;
    Double_t td = data_v[j].td;
    Double_t qave = data_v[j].qave;
    Double_t tave = data_v[j].tave;
    TVector3 pos = data_v[j].pos;

    fHIDX->Fill(id, pos.x());
    fHIDY->Fill(id, pos.y());
    fHIDZ->Fill(id, pos.z());
    fHIDQL->Fill(id, qu);
    fHIDQR->Fill(id, qd);
    fHIDTL->Fill(id, tu);
    fHIDTR->Fill(id, td);
    fHIDQAve->Fill(id, qave);
    fHIDTAve->Fill(id, tave);

    double energy = data_v[j].energy;
    double posx = pos[0]+ gRandom->Uniform(-6,6);
    double theta = atan(sqrt(posx*posx+pos[1]*pos[1])/pos[2]);
    fhenergy->Fill(energy);
    fhenergytheta->Fill(energy,theta);
    fhidenergy->Fill(id, energy);
    fhenergyqave->Fill(energy,qave);

    fHXTheta->Fill(posx, theta);

    double beamtanx = beammomentum.x()/beammomentum.z();
    double beamtany = beammomentum.y()/beammomentum.z();
    double tanx = posx/pos.z();
    double tany = pos.y()/pos.z();
    fHTanXTanY->Fill(tanx, tany);
    fHBeamTanXBeamTanY->Fill(-beamtanx, beamtany);
    fHBeamTanXTanX->Fill(-beamtanx, tanx);
    fHBeamTanYTanY->Fill(beamtany, tany);
    fHDTanXDTanY->Fill(-beamtanx-tanx, beamtany-tany);
    fHEnergyContami->Fill(energy);

    double r = sqrt(pos[0]*pos[0]+pos[1]*pos[1]+pos[2]*pos[2]);
    double TOF = r/beambeta/29.979;
    for(int i=0; i<1; ++i){
      fHTQ[i]->Fill(tave - TOF, qave);
      fHQT[i]->Fill(qave, tave - TOF);
    }

    fHXY[0]->Fill(pos[0],pos[1]);

    //    std::cout << beammomentum[0] << " " << data_v[j].p[0] << std::endl;
    fHBPPX->Fill(beammomentum[0],data_v[j].p[0]);
    fHBPPY->Fill(beammomentum[1],data_v[j].p[1]);
    fHBPPZ->Fill(beammomentum[2],data_v[j].p[2]);
    fHDPX->Fill((beammomentum[0]+data_v[j].p[0])/beamp);
    fHDPY->Fill((beammomentum[1]-data_v[j].p[1])/beamp);
    fHDPZ->Fill((beammomentum[2]-data_v[j].p[2])/beamp);
    fHDX->Fill(beamtanx*pos[2]+pos[0]);
    fHDY->Fill(beamtany*pos[2]-pos[1]);
    fHDT->Fill(data_v[j].pos[2]/(29.979*beambeta*beammomentum[2]/beamp)-data_v[j].time);
  }

  return 1;
}

bool TAnalysis14Be::vetolocal(std::vector<TAnalysisNebulaData>& data_v)
{
  std::vector<TAnalysisNebulaData> killer;
  for(int i=0; i<(int)data_v.size(); ++i){
    TAnalysisNebulaData data = data_v[i];

    if(data.mt == kVETO){
      if(data.id[2] == 0){
	return true;
      }else if(data.id[2] == 1){
	killer.push_back(data);
      }
    }
  }

  // kill veto
  for(int i=0; i<(int)data_v.size(); ++i){
    TAnalysisNebulaData data = data_v[i];

    if(data.mt == kVETO){
      data_v.erase(data_v.begin()+i);
      --i;
    }
  }

  for(int i=0; i<(int)data_v.size(); ++i){
    TAnalysisNebulaData data = data_v[i];

    if(data.id[2] == 1){
      for(int j=0; j<(int)killer.size(); ++j){
	double dx = data.pos[0]-killer[j].pos[0];
	double dy = data.pos[1]-killer[j].pos[1];
	double r = sqrt(dx*dx+dy*dy);
	if((r < 60) && 
	   (data.time - killer[j].time > 1) &&
	   (data.time - killer[j].time < 6)){
	  data_v.erase(data_v.begin()+i);
	  --i;
	  break;
        }
      }
    }
  }

  return false;
}
